"""Algebra utilities involved in the deduction procedure of fusion plasma physics formula.

Author: Wenyin Wei wenyin.wei@ipp.ac.cn

"""


def run_once(f):
    """Decorator to describe a function that only needs to run once and store the result in the memory in case of future reuse.

    From aaronasterling, StackOverflow https://stackoverflow.com/questions/4103773/efficient-way-of-having-a-function-only-execute-once-in-a-loop.

    Args:
        f (function): A function that only needs to run once
    """
    def wrapper(*args, **kwargs):
        if not wrapper.has_run:
            wrapper.has_run = True
            wrapper.result = f(*args, **kwargs)
            return wrapper.result
        else:
            return wrapper.result
    wrapper.has_run = False
    wrapper.result = None
    return wrapper


def find_singularities(expr, var):
    from sympy.core.power import Pow
    from sympy.core.mul import Mul
    from sympy.core.add import Add
    from sympy import solve, Eq
    from operator import or_
    from functools import reduce

    # recurse the function on each arg if the top function is an addition
    # e.g. 4 * sqrt(b) * sqrt(c) should also be regarded as a sqrt expression.
    if expr.func == Add or expr.func == Mul:
        args_singularities_ = [find_singularities(arg, var) for arg in expr.args]
        return reduce(or_, args_singularities_) # This trick from Alexander Klimenko, https://stackoverflow.com/questions/17429123/how-to-join-two-sets-in-one-line-without-using
    # Main Function 
    elif expr.func == Pow and expr.args[1].is_negative: # The second arg of Pow is a negative number if the term is a denominator
        return set(solve( Eq(expr.args[0], 0), var )) | find_singularities(expr.args[0], var)
    else:
        return set()


# def find_singularities_with_type(expr, var, n_or_d, part_of_addition:bool):
#     from sympy.core.power import Pow
#     from sympy.core.mul import Mul
#     from sympy.core.add import Add
#     from sympy import solve, Eq
#     from operator import or_
#     from functools import reduce

#     # recurse the function on each arg if the top function is an addition
#     # e.g. 4 * sqrt(b) * sqrt(c) should also be regarded as a sqrt expression.
#     if expr.func == Add:
#         args_singularities_ = [find_singularities_with_type(arg, var, n_or_d, part_of_addition=True) for arg in expr.args]
#         return reduce(or_, args_singularities_) # This trick from Alexander Klimenko, https://stackoverflow.com/questions/17429123/how-to-join-two-sets-in-one-line-without-using
#     if expr.func == Mul:
#         args_singularities_ = [find_singularities_with_type(arg, var, n_or_d, part_of_addition=False) for arg in expr.args]
#         return reduce(or_, args_singularities_) # This trick from Alexander Klimenko, https://stackoverflow.com/questions/17429123/how-to-join-two-sets-in-one-line-without-using
#     # Main Function 
#     elif expr.func == Pow and expr.args[1].is_negative: # The second arg of Pow is a negative number if the term is a denominator
#         if n_or_d == 'n':
#             singularities_ = [(sol, 'inf') for sol in solve( Eq(expr.args[0], 0), var )]
#         elif n_or_d == 'd' and not part_of_addition:
#             singularities_ = [(sol, 'zero') for sol in solve( Eq(expr.args[0], 0), var )]
#         elif n_or_d
#         return set(singularities_) | find_singularities_with_type(expr.args[0], var, not n_or_d)
#     else:
#         return set()

def find_zeroes(expr, var):
    """find all zero points of a sympy expression(var).

    Args:
        expr (sympy.expr): The sympy expression for which the zero points are requested.
        var (sympy.expr): The variable(s) of the expression

    Returns:
        set: a set of floating numbers which are the variable values as the zero points of the expression.

    Note:
        The most naive way to find zeros is to use sympy.solve(sympy.Eq(expr, 0), var). 
        However, if we use the following expr as an example,
        (1/(x-2)+2) / (1 + 1/x + 1/(x-1))
        sympy.solve can only find the x=3/2 zero point, without the x=0 and x=1. 

        Our custom function is more smart, it can find more complete solutions, {0,1,3/2} in the above example.
        I admit that this find_zeroes function has not been comprehensive due to its simple implementation, yet smart enough for the code author's application. 
        
    """
    from sympy.core.power import Pow
    from sympy.core.mul import Mul
    from sympy.core.add import Add
    from sympy import solve, Eq, cancel, fraction
    from operator import or_
    from functools import reduce

    # recurse the function on each arg if the top function is an addition
    # e.g. 4 * sqrt(b) * sqrt(c) should also be regarded as a sqrt expression.
    expr = cancel(expr)
    n, d = fraction(expr)
    zeros_from_numerator = set(solve( Eq(n, 0), var ))
    return zeros_from_numerator


def split_logspace(start, stop, breakpoints:set, num=50, 
                   endpoint=True , base=10.0, dtype=None):
    from numpy import logspace
    start_end_pair = sorted({point for point in breakpoints if point > start and point < stop})
    start_end_pair = [start] + start_end_pair + [stop]
    result = []
    for i in range(len(start_end_pair)-1):
        subnum = int( (start_end_pair[i+1] - start_end_pair[i]) * num / (stop - start) )
        if endpoint:
            result.append(
                logspace(start_end_pair[i], start_end_pair[i+1], subnum, endpoint, base, dtype)
            )
        else:
            subspacing = (start_end_pair[i+1] - start_end_pair[i]) / subnum
            result.append(
                logspace(start_end_pair[i] + subspacing, start_end_pair[i+1], subnum, endpoint, base, dtype)
            )
    return result

    


"""Algebra utilities involved in the deduction procedure of fusion plasma physics formula.

Author: Wenyin Wei wenyin.wei@ipp.ac.cn

"""

from sympy import Array as _Array
from sympy import LeviCivita as _LeviCivita
from sympy import tensorcontraction as _tcontract
from sympy import tensorproduct as _tprod

# t_LC means a LeviCivita *t*ensor.
t_LC = _Array([[[_LeviCivita(i1+1,i2+1,i3+1) for i3 in range(3)] for i2 in range(3)] for i1 in range(3)])

def cross(A, B):
    _tcontract(_tprod(A, B, t_LC),(0,2),(1,3))

def dot(A, B):
    # TODO
    pass 

def m_A_x(A):
    """returns a matrix Array equivalent with $\vec{A} \times$ operation, where $\vec{A}$ is a vector while $\times$ is cross product.  

    Args:
        A (sympy.Array): A should be a vector

    Returns:
        sympy.Array: A Matrix equivalent with $\vec{A} \times$ operation.
    """
    return _tcontract(_tprod(t_LC, A), (1,3))



"""Algebra utilities involved in the deduction procedure of fusion plasma physics formula.

Author: Wenyin Wei wenyin.wei@ipp.ac.cn

"""

def draw_discontinuable_expr(exprs, var, varlim, exprlim=None, var_is_yaxis=False, num=50, var_sample_scale='linear', fig=None, ax=None, labels=None, list_subkwarg=None):
    from numpy import linspace, logspace, logical_or, log10, nan
    from sympy import lambdify
    
    if ax is None:
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        
    for i, expr in enumerate(exprs):
        if var_sample_scale == 'linear':
            range_ = linspace(varlim[0], varlim[1], num=num)
        elif var_sample_scale == 'log':
            range_ = logspace(log10(varlim[0]), log10(varlim[1]), num=num)
        else:
            raise ValueError("Choose either linear or log scale.")
        
        values_ = lambdify(var, expr, 'numpy')(range_)
        if exprlim:
            if exprlim[0]:
                values_[values_ < exprlim[0]] = nan
            if exprlim[1]:
                values_[values_ > exprlim[1]] = nan

        if not var_is_yaxis:
            plot_func_args = (range_, values_)
        else:
            plot_func_args = (values_, range_)

        if list_subkwarg and list_subkwarg[i]:
            plot_func_kwargs = list_subkwarg[i]
        else: 
            plot_func_kwargs = dict()
        if labels is not None: 
            plot_func_kwargs['label'] = labels[i]

        ax.plot(*plot_func_args, **plot_func_kwargs)
        
    return fig, ax

def add_line_with_slope(ax, k:float, num=50, **kwargs):
    from numpy import linspace
    lower, upper = ax.get_xbound()
    x_range_ = linspace(lower, upper, num=num)
    y_range_ = k * x_range_ 
    ax.plot(x_range_, y_range_, **kwargs)


"""Algebra utilities involved in the deduction procedure of fusion plasma physics formula.

Author: Wenyin Wei wenyin.wei@ipp.ac.cn

"""

from scipy.constants import e as _e
from scipy.constants import m_e as _m_e
from scipy.constants import m_n as _m_n
from scipy.constants import m_p as _m_p
from scipy.constants import epsilon_0 as _epsilon_0
from scipy.constants import physical_constants as _physical_constants
from sympy import sqrt as _sqrt

def omega_cs(q_e, m, B):
    omega_cs_ = abs(q_e) * _e * B / m
    return omega_cs_, 'rad s^-1', None
def omega_ps(n_0, q_e, m, B):
    omega_ps_ = (q_e*_e) * _sqrt( n_0 / (_epsilon_0 * m) ) # Split the square of q_e * e, avoid it being too little
    return omega_ps_, 'rad s^-1', None

def omega_ce(B):
    return omega_cs(-1, _m_e, B)
def omega_pe(n_0, B):
    return omega_ps(n_0, -1, _m_e, B)

def omega_cp(B):
    return omega_cs(+1, _m_p, B)
def omega_pp(n_0, B):
    return omega_ps(n_0, +1, _m_p, B)

def mass_electron():
    return _physical_constants['electron mass']
def mass_proton():
    return _physical_constants['proton mass']
def mass_deuteron():
    return _physical_constants['deuteron mass']
def mass_triton():
    return _physical_constants['triton mass']

def omega_cD(B):
    m_D = _physical_constants['deuteron mass'][0]
    return omega_cs(+1, m_D, B)
def omega_pD(n_0, B):
    m_D = _physical_constants['deuteron mass'][0]
    return omega_ps(n_0, +1, m_D, B)

def omega_cT(B):
    m_T = _physical_constants['triton mass'][0]
    return omega_cs(+1, m_T, B)
def omega_pT(n_0, B):
    m_T = _physical_constants['triton mass'][0]
    return omega_ps(n_0, +1, m_T, B)


"""Algebra utilities involved in the deduction procedure of fusion plasma physics formula.

Author: Wenyin Wei wenyin.wei@ipp.ac.cn

"""

from sympy import Symbol as _Symbol 
from sympy import symbols as _symbols
from sympy import Array as _Array
from sympy import Matrix as _Matrix
from sympy import LeviCivita as _LeviCivita
from sympy import tensorcontraction as _tcontract
from sympy import tensorproduct as _tprod
from fusystem.algebra.utility import run_once

class deduction_dispersion_in_plasma:

    k_0 = _Symbol('k_0', positive=True)
    theta = _Symbol('theta', nonnegative=True)
    N = _Symbol('N')
    k_x, k_y, k_z, k = _symbols('k_x, k_y, k_z, k', real=True); vk = _Array([k_x, k_y, k_z])
    omega, omega_pe, omega_ce, omega_pi, omega_ci = _symbols(
        'omega, \omega_{pe}, \omega_{ce}, \omega_{pi}, \omega_{ci}', negative=False)
    E_x, E_y, E_z = _symbols('E_x, E_y, E_z'); vE = _Array([E_x, E_y, E_z])
    kappa_para, kappa_times, kappa_perp = _symbols('kappa_\parallel, \kappa_{\\times}, kappa_\perp', real=True)
    B = _Symbol('B')
    n_e, n_i = _symbols('n_e, n_i')

    @classmethod
    def relative_dielectric_tensor(cls):
        from sympy import I
        return _Matrix([
            [cls.kappa_perp,    -I*cls.kappa_times, 0],
            [I*cls.kappa_times, cls.kappa_perp,     0],
            [0,                 0,     cls.kappa_para]])

    @classmethod
    def tensor_in_wave_equation(cls):
        from fusystem.algebra.tensor import m_A_x
        m_vk_x = m_A_x(cls.vk)
        return m_vk_x.tomatrix() * m_vk_x.tomatrix() + cls.k_0**2 * cls.relative_dielectric_tensor() 

    @classmethod
    def subs_k_component_with_k_theta(cls, expr):
        from sympy import sin, cos
        expr = expr.subs(cls.k_y, 0)
        expr = expr.subs(cls.k_x, cls.k* sin(cls.theta)).subs(cls.k_z, cls.k*cos(cls.theta))
        return expr

    @classmethod
    def subs_k_with_k_0_and_N(cls, expr):
        return expr.subs(cls.k, cls.k_0 * cls.N)

    @classmethod
    def set_theta_value(cls, expr, theta_value):
        return expr.subs(cls.theta, theta_value)

    # @classmethod
    # def solve_N2_with_theta(cls, theta_value):
    #     N2_sol_theta_ = [sol.subs(theta, theta_value).simplify() for sol in N2_sol]
    #     return list(fuquad.simplify_quadratic_sols(N2_sol_theta_[0], N2_sol_theta_[1]))

    @classmethod
    @run_once
    def solve_N2_with_theta_and_kappa_component(cls):
        from sympy import solve, Eq
        tensor_det_in_wave_equation = cls.tensor_in_wave_equation().det()
        tensor_det_in_wave_equation_with_k_0_and_N = \
            cls.subs_k_with_k_0_and_N(
                cls.subs_k_component_with_k_theta(
                    tensor_det_in_wave_equation))
        N2_sol = solve(Eq(tensor_det_in_wave_equation_with_k_0_and_N.simplify(), 0), (cls.N**2))
        return [sol.simplify() for sol in N2_sol]
    
    @classmethod
    def solve_N2_with_specific_theta_and_kappa_component(cls, theta_value):
        from fusystem.algebra.quadratic import simplify_quadratic_sols
        N2_with_specific_theta_and_kappa_component_ = [
                cls.set_theta_value(sol, theta_value) 
            for sol in cls.solve_N2_with_theta_and_kappa_component()]
        return list(simplify_quadratic_sols(
            N2_with_specific_theta_and_kappa_component_[0],
            N2_with_specific_theta_and_kappa_component_[1]))

    @classmethod
    def subs_kappa_with_omega(cls, expr, plasma_type='electron'):
        """Substitute kappa_components with omega components.

        Args:
            expr (sympy.expr): [description]
            plasma_type (str, optional): [description]. Defaults to 'electron'.

        Raises:
            ValueError: [description]

        Returns:
            [type]: [description]

        DevNote:
            Do not try to simplify the expression! The expression can be so complicated that a lot of time would be wasted to get a nonsense.

        """
        f = lambda a,b,c: a**2 / (b**2 - c**2)
        if plasma_type == 'electron':
            o, o_pe, o_ce = cls.omega, cls.omega_pe, cls.omega_ce
            expr = expr\
                .subs(cls.kappa_perp, 1 - f(o_pe, o, o_ce))\
                .subs(cls.kappa_times, (o_ce / o) * f(o_pe, o, o_ce))\
                .subs(cls.kappa_para, 1 - o_pe**2 / o**2 ) 
        elif plasma_type == 'electron+ion':
            o, o_pe, o_ce, o_pi, o_ci = cls.omega, cls.omega_pe, cls.omega_ce, cls.omega_pi, cls.omega_ci
            expr = expr\
                .subs(cls.kappa_perp, 
                    1 - f(o_pe, o, o_ce) - f(o_pi, o, o_ci))\
                .subs(cls.kappa_times, 
                    o_ce/o * f(o_pe, o, o_ce) - o_ci/o * f(o_pi, o, o_ci))\
                .subs(cls.kappa_para,
                    1 - o_pe**2/o**2)
        else:
            raise ValueError("Unprepared plasma type")
        return expr




"""Algebra utilities involved in the deduction procedure of fusion plasma physics formula.

Author: Wenyin Wei wenyin.wei@ipp.ac.cn

"""

################################################################################
# Judge whether a term could be a sqrted term, squared term or an absoluted term.
################################################################################

def is_sqrt(expr):
    from sympy.core.power import Pow
    from sympy.core.mul import Mul
    from sympy.core.numbers import Half
    
    is_sqrt_ = False # defaults to be False
    expr = expr.factor()
    # recurse the function on each arg if the top function is a multiplication 
    # e.g. 4 * sqrt(b) * sqrt(c) should also be regarded as a sqrt expression.
    if expr.func == Mul: 
        args_is_sqrt = [is_sqrt(arg) for arg in expr.args]
        return True if all(args_is_sqrt) else False
    # Main Function 
    elif expr.func == Pow: 
        if isinstance(expr.args[1], Half): # The second arg of Pow is a Half number if the term is a sqrt
            is_sqrt_ = True
    elif not expr.is_negative:
        is_sqrt_ = True
    return is_sqrt_

def is_square(expr):
    from sympy.core.power import Pow
    from sympy.core.mul import Mul
    
    is_square_ = False # defaults to be False
    expr = expr.factor()
    # recurse the function on each arg if the top function is a multiplication 
    # e.g. 4 * b^4 * b^6 * c^2 should also be regarded as a square expression.
    if expr.func == Mul: 
        args_is_square = [is_square(arg) for arg in expr.args]
        return True if all(args_is_square) else False
    # Main Function 
    elif expr.func == Pow:
        base, exponent = expr.args
        if exponent % 2 == 0: # The second arg of Pow is an even number if the term is a square
            is_square_ = True
    elif not expr.is_negative:
        is_square_ = True
    return is_square_

def is_abs(expr):
    from sympy.core.mul import Mul
    from sympy import Abs
    
    is_abs_ = False # defaults to be False
    expr = expr.factor()
    # recurse the function on each arg if the top function is a multiplication 
    # e.g. 4 * Abs(b) * Abs(c) should also be regarded as a abs expression.
    if expr.func == Mul: 
        args_is_abs = [is_abs(arg) for arg in expr.args]
        return True if all(args_is_abs) else False
    # Main Function 
    elif expr.func == Abs: 
        is_abs_ = True
    elif not expr.is_negative:
        is_abs_ = True 
    return is_abs_

def signed_sqrt(expr):
    """Signed sqrt operator

    Args:
        expr (sympy.expr): sympy expression

    Returns:
        sympy.expr: A simplified expression
    """
    from functools import reduce
    from sympy.core.power import Pow
    from sympy.core.mul import Mul
    from sympy import sqrt

    expr = expr.factor()
    # recurse the function on each arg if the top function is a multiplication 
    # e.g. signed_sqrt( 4 * b^2 ) == 2 * b 
    if expr.func == Mul: 
        args_signed_sqrt = [signed_sqrt(arg) for arg in expr.args]
        return reduce(Mul, args_signed_sqrt)
    elif expr.func == Pow:
        base, exponent = expr.args
        if exponent.is_even:
            return base**(exponent/2)
    return sqrt(expr)
        
def are_quadratic_sols(expr1, expr2):
    # Suppose the function is a form of -b +/- sqrt(Delta) / 2a, where Delta = b^2 - 4ac
    
    expr_sum  = expr1 + expr2 # -b/a
    expr_diff = expr1 - expr2 # +/- sqrt(Delta)/a
    expr_prod = expr1 * expr2 # c/a
    # b^2/a^2 - 4 c/a == (b^2 - 4ac) / a^2 
    lhs = (expr_sum ** 2 - 4 * expr_prod).simplify()
    rhs = (expr_diff ** 2).simplify()
    lhs_minus_rhs = (lhs - rhs).simplify()
    return True if lhs_minus_rhs==0 else False

def simplify_quadratic_sols(expr1, expr2):
    if not are_quadratic_sols(expr1, expr2):
        raise ValueError("The input expr1, expr2 may not be a pair of solutions of a quadratic equation. Make sure they look like -b +/- sqrt(Delta) / 2a.")

    expr_sum  =(expr1 + expr2).simplify() # -b/a
    expr_diff =(expr1 - expr2).simplify() # +/- sqrt(Delta)/a
    
    signed_sqrt_Delta_on_a = signed_sqrt(expr_diff**2) # Still +/- sqrt(Delta)/a, but simplified
    
    simplified_expr1 = (expr_sum + signed_sqrt_Delta_on_a)/2
    simplified_expr2 = (expr_sum - signed_sqrt_Delta_on_a)/2
    simplified_expr1 = simplified_expr1.simplify()
    simplified_expr2 = simplified_expr2.simplify()
    return simplified_expr1, simplified_expr2




import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sinupy",
    version="0.0.1",
    author="Wenyin Wei",
    author_email="wenyin.wei@ipp.ac.cn",
    description="The sinupy package make it convenient to analyze the characteristic sinusoidal waves propagating in various kinds of medium. Typical waves in plasma have been given in the code to demonstrate how the package works and other users are free to follow the steps to research on their medium or charcteristic waves of interest.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/WenyinWei/sinupy",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)